#pragma once
#include "gte_instruction.hpp"

namespace JabyEngine {
    namespace GTE {
        static constexpr auto StackSize = 16;

        /*
            RotTrans

            Jaby: Can we use gte_stsv instead of gte_stlvnl for writing to a SVECTOR?
                  Do we have to use gte_stflg??
                  Look at: RotTransSV???

            Perform coordinate transformation using a rotation matrix
            input:  Input vector
            output: Output vector
            flag:   flag output
        */
        static void rot_trans(const SVECTOR& input, VECTOR& output, int32_t& flag) {
            ldv0(input);
            rt();
            stlvnl(output);
            stflg(flag);
        }

        /*
            SetRotMatrix

            Sets a 3x3 matrix m as a constant rotation matrix.
            matrix: The rotation matrix to set
        */
        static void set_rot_matrix(const MATRIX& matrix) {
            __asm__ volatile("lw   $12, 0(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("lw   $13, 4(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("ctc2 $12, $0"     :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("ctc2 $13, $1"     :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("lw   $12, 8(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("lw   $13, 12(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("lw   $14, 16(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $12, $2"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $13, $3"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $14, $4"     :: "r"(&matrix) : "$12", "$13", "$14");
        }

        /*
            GetRotMatrix

            Writes the current 3x3 constant rotation matrix to matrix
            (This doesn't require us to use memory clobber)
        */
        static void get_rot_matrix(MATRIX &matrix) {
            __asm__ volatile("cfc2 $12, $0"     :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("cfc2 $13, $1"     :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("sw   $12, 0(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("sw   $13, 4(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("cfc2 $12, $2"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("cfc2 $13, $3"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("cfc2 $14, $4"     :: "r"(&matrix) : "$12", "$13", "$14");            
            __asm__ volatile("sw   $12, 8(%0)"  :: "r"(&matrix) : "$12", "$13", "$14"); 
            __asm__ volatile("sw   $13, 12(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("sw   $14, 16(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
        }

        /*
            SetTransMatrix

            Sets a constant parallel transfer vector specified by m
        */
        static void set_trans_matrix(const MATRIX& matrix)  {
            __asm__ volatile("lw   $12, 20(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("lw   $13, 24(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $12, $5"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("lw   $14, 28(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $13, $6"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("ctc2 $14, $7"     :: "r"(&matrix) : "$12", "$13", "$14");
        }

        /*
            GetTransMatrix

            Writes the current constant parallel transfer vector to matrix
            (This doesn't require us to use memory clobber)
        */
        static void get_trans_matrix(MATRIX& matrix) {
            __asm__ volatile("cfc2 $14, $7"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("cfc2 $13, $6"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("sw   $14, 28(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("cfc2 $12, $5"     :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("sw   $13, 24(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
            __asm__ volatile("sw   $12, 20(%0)" :: "r"(&matrix) : "$12", "$13", "$14");
        }              

        /*
            MulMatrix0

            m0: first input
            m1: second input
            result: result of multiplication
            returns: result

            Multiplies two matrices m0 and m1.
            The function destroys the constant rotation matrix
        */
        MATRIX& multiply_matrix(const MATRIX& m0, const MATRIX& m1, MATRIX& result);

        /*
            matrix: first input

            Sets the 3x3 constant rotation matrix and the parallel transfer vector from input 
        */
        void set_matrix(const MATRIX& matrix);

        /*
            returns: current matrix

            Gets the current 3x3 constant rotation matrix and the parallel transfer vector
        */
        MATRIX get_matrix();

        /*
            matrix: optional input

            Pushes the current matrix (rotation and parallel) to an internal stack
            Optional: replaces current matrix (rotation and parallel) with input
        */
        void push_matrix();
        void push_matrix_and_set(const MATRIX& matrix);

        /*
            Restores the previous stored matrix (rotation and parallel)
        */
        MATRIX get_and_pop_matrix();
        void   pop_matrix();

        /*
            SetGeomOffset(ofx,ofy)

            Load GTE-offset.
        */
        static void set_geom_offset(int32_t off_x, int32_t off_y) {
            __asm__ volatile("sll  $12, %0, 16" :: "r"(off_x), "r"(off_y) : "$12", "$13");
            __asm__ volatile("sll  $13, %1, 16" :: "r"(off_x), "r"(off_y) : "$12", "$13");
            __asm__ volatile("ctc2 $12, $24"    :: "r"(off_x), "r"(off_y) : "$12", "$13");
            __asm__ volatile("ctc2 $13, $25"    :: "r"(off_x), "r"(off_y) : "$12", "$13");
        }

        /*
            SetGeomScreen(h)
            
            Load distance from viewpoint to screen.
        */
        static void set_geom_screen(int32_t h) {
            __asm__ volatile("ctc2 %0, $26" :: "r"(h));
        }
    }
}