#include "../../internal-include/GPU/gpu_internal.hpp"
#include <PSX/Timer/frame_timer.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.hpp>

namespace JabyEngine {
    namespace GPU {
        uint8_t Display :: current_id = 1; //< Setup will call exchange and set it to 0

        namespace internal {
            namespace DMA {
                #ifdef SUPPORT_PS3
                    uintptr_t MADR = 0;
                #endif // SUPPORT_PS3
            }

            static SysCall::InterruptVerifierResult interrupt_verifier();
            static uint32_t interrupt_handler(uint32_t);

            static uint8_t vsync_counter = 0;

            SysCall::InterrupCallback callback = {
                .next              = nullptr,
                .handler_function  = interrupt_handler,
                .verifier_function = interrupt_verifier
            };

            static SysCall::InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::VBlank)) {
                    return SysCall::InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return SysCall::InterruptVerifierResult::SkipHandler;
                }
            }

            static uint32_t interrupt_handler(uint32_t) {
                vsync_counter++;
                MasterTime::value++;

                Interrupt::ack_irq(Interrupt::VBlank);
                SysCall::ReturnFromException();
                __builtin_unreachable();
            }

            uint32_t Display :: exchange_buffer_and_display() {
                static constexpr uint16_t TexturePageHeight = 256;

                const uint16_t draw_area_y = (TexturePageHeight*PublicDisplay::current_id);

                GPU::internal::set_draw_area(GPU::PositionU16::create(0, draw_area_y));
                PublicDisplay::current_id ^= 1;
                GPU_IO::GP1.write(GPU_IO::Command::DisplayArea(GPU::PositionU16::create(0, static_cast<uint16_t>((TexturePageHeight*PublicDisplay::current_id)))));
                return draw_area_y;
            }

            void wait_vsync(uint8_t syncs) {
                volatile auto& vsync_count = reinterpret_cast<volatile uint8_t&>(vsync_counter);
                const uint8_t  dst_value   = vsync_count + syncs;

                while(vsync_count != dst_value);
            }

            void render(const uint32_t* data, size_t words) {
                wait_ready_for_CMD();
                for(size_t n = 0; n < words; n++) {
                    GPU_IO::GP0.write({data[n]});
                }
            }

            void render_dma(const uint32_t* data) {
                // DPCR already enabled
                DMA_IO::GPU.wait();
                GPU_IO::GP1.write(GPU_IO::Command::DMADirection(GPU_IO::DMADirection::CPU2GPU));
                DMA_IO::GPU.set_adr(reinterpret_cast<uintptr_t>(data));
                DMA_IO::GPU.block_ctrl.write(DMA_IO::BCR::SyncMode2::for_gpu_cmd());

                wait_ready_for_CMD();
                DMA_IO::GPU.channel_ctrl.write(DMA_IO::CHCHR::StartGPULinked());
            }
        }

        void Display :: set_offset(int16_t x, int16_t y) {
            x += internal::Display::DisplayRange.x;
            y += internal::Display::DisplayRange.y;

            GPU_IO::GP1.write(GPU_IO::Command::HorizontalDisplayRange((x << 3), (x + Display::Width) << 3));
            GPU_IO::GP1.write(GPU_IO::Command::VerticalDisplayRange(y,         y + Display::Height));
        }

        uint8_t swap_buffers_vsync(uint8_t syncs, bool clear_screen) {
            // Waits for finish FIFO
            internal::wait_ready_for_CMD();

            internal::wait_vsync(syncs);
            const int16_t draw_offset_y = internal::Display::exchange_buffer_and_display();
            internal::set_draw_offset(GPU::PositionI16::create(0, draw_offset_y));
            if(clear_screen) {
                internal::quick_fill_fast(Color24::Black(), AreaU16::create(0, static_cast<uint16_t>(draw_offset_y), Display::Width, Display::Height));
            }
            return Display::current_id;
        }
    }
}